// File:  keyboard.v
// Date:  01/01/2005
// Name:  Eric Crabill
// 
// This is the top level design for the
// EE178 Lab #5 assignment.

// The `timescale directive specifies what
// the simulation time units are (1 ns here)
// and what the simulator timestep should be
// (1 ps here).

`timescale 1 ns / 1 ps

module keyboard (ps2_clk, ps2_dat, rd_kbrd, rd_dbug, clk, rst, 
                 rd_data, data_present, data_half, data_full);

  input ps2_clk;  // must be pulled up at pin
  input ps2_dat;  // must be pulled up at pin

  input rd_kbrd;
  input rd_dbug;
  input clk;
  input rst;

  output [7:0] rd_data;
  output data_present;
  output data_half;
  output data_full;

  // In no particular order, start of by generating
  // the FIFO read control signal.  There are two
  // ways for the FIFO to be read.  One is using a
  // synchronous input, rd_kbrd.  The other is using
  // an asynchronous pushbutton input, rd_dbug.  For
  // stand-alone testing, rd_dbug would be connected
  // to a pushbutton and rd_kbrd would be tied to 0.
  // For integration in a synchronous system, you
  // would want rd_dbug tied to 0 and rd_kbrd tied to
  // some signal that causes the FIFO to be read.

  wire read;
  reg dbug_d1, dbug_d2;
  reg [19:0] dbug_ctr;
  reg dbug_det;

  always @(posedge clk or posedge rst)
  begin
    if (rst)
    begin
      // just reset everything
      dbug_d1 <= 1'b0;
      dbug_d2 <= 1'b0;
      dbug_ctr <= 20'h00000;
      dbug_det <= 1'b0;
    end
    else
    begin
      // next two lines are synchronizer
      dbug_d1 <= rd_dbug;
      dbug_d2 <= dbug_d1;
      if (dbug_d2)
      begin
        // button is pressed
        // count up to 20'hfffff and stop
        // if we get to 20'hffffe then
        // signal 1 cycle debounce detect
        if (dbug_ctr != 20'hfffff) dbug_ctr <= dbug_ctr + 20'h00001;
        dbug_det <= (dbug_ctr == 20'hffffe); 
      end
      else
      begin
        // button not pressed
        // clear everything
        dbug_ctr <= 20'h00000;
        dbug_det <= 1'b0;
      end
    end
  end

  assign read = rd_kbrd | dbug_det;

  // Perform the same type of operation as
  // done above with the ps2_clk signal.
  // This signal has slow transition rates
  // (particularly when rising due to the
  // passive pull-up).  The idea is to keep
  // noise from looking like edges on ps2_clk.
  // Also synchronize ps2_dat, although you
  // can argue this is not required if the
  // keyboard is following the ps2 protocol.
   
  wire ps2_data, ps2_edge;
  reg data_d1, data_d2;
  reg edge_d1, edge_d2;
  reg [3:0] edge_ctr;
  reg edge_det;

  always @(posedge clk or posedge rst)
  begin
    if (rst)
    begin
      // just reset everything
      data_d1 <= 1'b1;
      data_d2 <= 1'b1;
      edge_d1 <= 1'b1;
      edge_d2 <= 1'b1;
      edge_ctr <= 4'h0;
      edge_det <= 1'b0;
    end
    else
    begin
      // next four lines are synchronizer
      data_d1 <= ps2_dat;
      data_d2 <= data_d1;
      edge_d1 <= ps2_clk;
      edge_d2 <= edge_d1;
      if (edge_d2)
      begin
        // clock signal is high
        // clear everything
        edge_ctr <= 4'h0;
        edge_det <= 1'b0;
      end
      else
      begin
        // clock signal is low
        // count up to 4'hf and stop
        // if we get to 4'he then
        // signal 1 cycle edge detect
        if (edge_ctr != 4'hf) edge_ctr <= edge_ctr + 4'h1;
        edge_det <= (edge_ctr == 4'he); 
      end
    end
  end

  assign ps2_edge = edge_det;
  assign ps2_data = data_d2;

  // Implement an 11-bit shift register that has
  // a single shift control and data input.  This
  // shift register also has two outputs; one is
  // the 8-bit received data and the other is a
  // status signal that indicates the start, stop,
  // and parity check out okay.

  wire shift, okay;
  reg [10:0] shifter;
  wire [7:0] data_in;

  always @(posedge clk or posedge rst)
  begin
    if (rst) shifter <= 11'b00000000000;
    else if (shift) shifter <= {ps2_data, shifter[10:1]};
  end
  
  assign data_in = shifter[8:1];
  assign okay = !shifter[0] && shifter[10] && ^shifter[9:1];

  // Implement a time-out counter.  Each time a bit
  // is shifted, start counting.  After 100 usec or
  // so, signal that the timer has expired.  The
  // expired signal is generated with combinational
  // logic so that in the next cycle after a shift,
  // the expired signal will be deasserted.  This
  // ensures the FSM can start looking at expired
  // immediately after a shift state.  The idea is
  // that the FSM, while waiting for the next bit,
  // can abort if this timer expires.  Hopefully
  // this will handle cases of surprise keyboard
  // removal during a transmission, and cases of
  // controller resets during transmission.

  wire expired;
  reg [11:0] timeout_ctr;

  always @(posedge clk or posedge rst)
  begin
    if (rst) timeout_ctr <= 12'h000;
    else if (shift) timeout_ctr <= 12'h000;
    else if (timeout_ctr != 12'hfff) timeout_ctr <= timeout_ctr + 12'h001;
  end

  assign expired = (timeout_ctr == 12'hffe);

  // Instantiate the fifo_16x8 module.  This acts
  // as a 16-deep scan code buffer from the keyboard
  // so that whoever's reading the scan codes does not
  // need to pick them up immediately upon reception.

  wire write;

  fifo_16x8 my_fifo (
    .data_in(data_in),
    .data_out(rd_data),
    .rst(rst),
    .write(write),
    .read(read),
    .full(data_full),
    .half_full(data_half),
    .data_present(data_present),
    .clk(clk));

  // Instantiate the FSM that controls the datapath.
  // You will have to design this for yourself.

  fsm my_fsm (
    .shift(shift),
    .write(write),
    .expired(expired),
    .okay(okay),
    .ps2_edge(ps2_edge),
    .clk(clk),
    .rst(rst));

endmodule